"""A raycaster gadget."""

import numpy as np

from ..array_types import Coords, ULong1D, ULong2D
from ..text_tools import _text_to_cells, _write_cells_to_canvas, cell_dtype
from ._raycasting import text_cast_rays
from .text import (
    Cell0D,
    Cell2D,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    Text,
)

__all__ = ["Point", "Size", "TextRaycaster"]


class TextRaycaster(Text):
    r"""
    A raycaster gadget that renders with text.

    ``caster_map`` should not contain a value greater than the length of
    ``wall_textures``. A non-zero value ``N`` in ``caster_map`` represents a wall with
    texture ``wall_textures[N - 1]``.

    The integer arrays in ``wall_textures`` determine how walls are shaded. With low
    values darker and high values lighter.

    ``sprite_indexes`` should not contain a value greater than or equal to the length of
    ``sprite_textures``. A value ``N`` in ``sprite_indexes`` represents a sprite with
    texture ``sprite_textures[N]``. ``sprite_coords`` and ``sprite_indexes`` must be the
    same length.

    It's convention in ``batgrl`` to refer to RGBA arrays as textures. ``wall_textures``
    and ``sprite_textures`` do not contain RGBA arrays, but the names are kept to be
    consistent with the graphics raycaster.

    Parameters
    ----------
    caster_map : ULong2D
        The raycaster map.
    wall_textures : list[ULong2D]
        Textures for walls.
    camera_coord : tuple[float, float], default: (0.0, 0.0)
        The camera's position.
    camera_angle : float, default: 0.0
        The camera's angle.
    camera_fov : float, default: 0.66
        The camera's field-of-view. Somewhere between 0-1.
    sprite_coords : Coords | None, default: None
        Positions of sprites.
    sprite_indexes : ULong1D | None, default: None
        Texture indexes of sprites.
    sprite_textures : list[str] | None, default: None
        Textures for sprites.
    ascii_map : str, default: " .,:;<+*LtCa4U80dQM@"
        Dark to bright ascii characters for shading.
    default_cell : Cell0D | str, default: " "
        Default cell of text canvas.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    caster_map : ULong2D
        The raycaster map.
    wall_textures : list[ULong2D]
        Textures for walls.
    camera_coord : tuple[float, float]
        The camera's position.
    camera_angle : float
        The camera's angle.
    camera_fov : float
        The camera's field-of-view. Somewhere between 0-1.
    sprite_coords : Coords
        Positions of sprites.
    sprite_indexes : ULong1D
        Texture indexes of sprites.
    sprite_textures : list[Cell2D]
        Textures for sprites.
    ascii_map : str
        Dark to bright ascii characters for shading.
    canvas : Cell2D
        The array of characters for the gadget.
    chars : Unicode2D
        Return a view of the ords field of the canvas as 1-character unicode strings.
    default_cell : Cell0D
        Default cell of text canvas.
    default_fg_color : Color
        Foreground color of default cell.
    default_bg_color : Color
        Background color of default cell.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    cast_rays()
        Update canvas by casting all rays.
    add_border(style="light", ...)
        Add a border to the gadget.
    add_syntax_highlighting(lexer, style)
        Add syntax highlighting to current text in canvas.
    add_str(str, pos, ...)
        Add a single line of text to the canvas.
    set_text(text, ...)
        Resize gadget to fit text, erase canvas, then fill canvas with text.
    clear()
        Fill canvas with default cell.
    shift(n=1)
        Shift content in canvas up (or down in case of negative `n`).
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        caster_map: ULong2D,
        wall_textures: list[ULong2D],
        camera_coord: tuple[float, float] = (0.0, 0.0),
        camera_angle: float = 0.0,
        camera_fov: float = 0.66,
        sprite_coords: Coords | None = None,
        sprite_indexes: ULong1D | None = None,
        sprite_textures: list[str] | None = None,
        ascii_map: str = " .,:;<+*LtCa4U80dQM@",
        default_cell: Cell0D | str = " ",
        alpha: float = 1.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            default_cell=default_cell,
            alpha=alpha,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.caster_map = caster_map
        """The raycaster map."""
        self.wall_textures = wall_textures
        """Textures for walls."""
        self.camera_coord = camera_coord
        """The camera's position."""
        self.camera_angle = camera_angle
        """The camera's angle."""
        self.camera_fov = camera_fov
        """The camera's field-of-view. Somewhere between 0-1."""

        self.sprite_coords: Coords
        """Positions of sprites."""
        if sprite_coords is None:
            self.sprite_coords = np.empty((0, 2), np.float64)
        else:
            self.sprite_coords = sprite_coords

        self.sprite_indexes: ULong1D
        """Texture indexes of sprites."""
        if sprite_indexes is None:
            self.sprite_indexes = np.empty(0, np.uint8)
        else:
            self.sprite_indexes = sprite_indexes

        self.sprite_textures: list[Cell2D] = []
        """Textures for sprites."""
        if sprite_textures is not None:
            for texture in sprite_textures:
                spr_size, lines = _text_to_cells(texture)
                canvas = np.empty(spr_size, cell_dtype)
                _write_cells_to_canvas(
                    lines, canvas, self.default_fg_color, self.default_bg_color
                )
                self.sprite_textures.append(canvas)

        self.ascii_map = ascii_map
        """Dark to bright ascii characters for shading."""

        self.on_size()

    @property
    def ascii_map(self) -> str:
        """Dark to bright ascii characters for shading."""
        return "".join(chr(i) for i in self._ascii_map)

    @ascii_map.setter
    def ascii_map(self, ascii_map: str):
        self._ascii_map = np.array([ord(char) for char in ascii_map], np.uint32)

    def cast_rays(self):
        """Update canvas by casting all rays."""
        self.clear()
        text_cast_rays(
            self.canvas,
            self.caster_map,
            self.camera_coord,
            self.camera_angle,
            self.camera_fov,
            self.wall_textures,
            self.sprite_indexes,
            self.sprite_coords,
            self.sprite_textures,
            self._ascii_map,
        )



"""Functions for blending colors and creating color gradients."""

from math import sin, tau
from typing import cast

import numpy as np

from ..geometry import clamp, lerp, normalize
from ..geometry.easings import EASINGS, Easing
from .color_types import AColor, Color

__all__ = ["darken_only", "gradient", "lerp_colors", "lighten_only", "rainbow_gradient"]


def darken_only(a: Color, b: Color) -> Color:
    """
    Return a color that is the minimum of each channel in ``a`` and ``b``.

    Parameters
    ----------
    a : Color
        A color.
    b : Color
        A color.

    Returns
    -------
    Color
        A color with smallest components of ``a`` and ``b``.
    """
    color = (min(c1, c2) for c1, c2 in zip(a, b))
    return Color(*color)


def lighten_only(a: Color, b: Color) -> Color:
    """
    Return a color that is the maximum of each channel in ``a`` and ``b``.

    Parameters
    ----------
    a : Color
        A color.
    b : Color
        A color.

    Returns
    -------
    Color
        A color with largest components of ``a`` and ``b``.
    """
    color = (max(c1, c2) for c1, c2 in zip(a, b))
    return Color(*color)


def lerp_colors[T: (Color, AColor, tuple[int, ...])](a: T, b: T, p: float) -> T:
    """
    Linear interpolation from ``a`` to ``b`` with proportion ``p``.

    Parameters
    ----------
    a : (Color, AColor, tuple[int, ...])
        A color.
    b : (Color, AColor, tuple[int, ...])
        A color.
    p : float
        Proportion from a to b.

    Returns
    -------
    (Color, AColor, tuple[int, ...])
        The linear interpolation of ``a`` and ``b``.
    """
    color = (clamp(round(lerp(c1, c2, p)), 0, 255) for c1, c2 in zip(a, b))
    if isinstance(a, (Color, AColor)):
        return type(a)(*color)
    return cast(T, tuple(color))


def gradient[T: (Color, AColor, tuple[int, ...])](
    *colors: T, n: int, easing: Easing = "linear"
) -> list[T]:
    r"""
    Return a smooth gradient of length ``n`` between all colors in ``colors``.

    Parameters
    ----------
    \*colors : (Color, AColor, tuple[int, ...])
        Pairwise colors determine start and end of each subgradient.
    n : int
        Length of gradient. Must be equal to or larger than ``colors``.
    easing : Easing, default: "linear"
        Easing applied to entire gradient.

    Returns
    -------
    list[(Color, AColor, tuple[int, ...])]
        A smooth gradient between all colors in ``colors``.
    """
    ncolors = len(colors)
    if n < ncolors:
        raise ValueError(f"gradient too small to contain all colors ({n=})")
    if ncolors == 1:
        return [colors[0]] * n

    ease = EASINGS[easing]
    pcolors: list[float] = np.linspace(0.0, 1.0, ncolors).tolist()
    eased_pcolors = [ease(p) for p in pcolors]

    i = 1
    gradient: list[T] = []
    for p in np.linspace(0.0, 1.0, n).tolist():
        while p > pcolors[i]:
            i += 1
        gradient.append(
            lerp_colors(
                colors[i - 1],
                colors[i],
                normalize(ease(p), eased_pcolors[i - 1], eased_pcolors[i]),
            )
        )
    return gradient


def rainbow_gradient(n: int, *, alpha: int | None = None) -> list[Color] | list[AColor]:
    """
    Return a rainbow gradient of ``n`` colors.

    Parameters
    ----------
    n : int
        Number of colors in gradient.
    alpha : int | None, default: None
        If ``alpha`` is not given, gradient colors will have no alpha channel.
        Otherwise, the color's alpha channel is given by ``alpha``.

    Returns
    -------
    list[Color | AColor]
        A rainbow gradient of colors.
    """
    theta = tau / n
    offsets: list[float] = [0, tau / 3, 2 * tau / 3]

    def color(i: int):
        return Color(*(int(sin(i * theta + offset) * 127 + 128) for offset in offsets))

    if alpha is None:
        return [color(i) for i in range(n)]

    return [AColor(*color(i), alpha) for i in range(n)]



"""
Functions and classes for determining gadget regions.

Notes
-----
A gadget's region is calculated as a first step in compositing to determine its visible
area in the terminal.

Let's say we have the following gadgets:

.. code-block:: text

    +--------+------+---------+
    |        |  B   |         |
    |        +------+         |
    |                         |
    |        A             +-------+
    |                      |   C   |
    +----------------------+-------+

And we want to represent the visible region of ``A``:

.. code-block:: text

    +--------+      +---------+
    |        |      |         |
    |        +------+         |
    |                         |
    |                      +--+
    |                      |
    +----------------------+

One method is to divide the area into a series of mutually exclusive horizontal bands:

.. code-block:: text

    +--------+      +---------+
    | a      | b    | c       | d    - Top band with walls at a, b, c, d
    |--------+------+---------+
    | e                       | f    - Middle band with walls at e, f
    +----------------------+--+
    | g                    | h       - Bottom band with walls at g, h
    +----------------------+

Walls are the x-coordinates of the rects in a band. Two contiguous walls indicate a new
rect. Bands are a sorted list of walls with each band having a top y-coordinate and
bottom y-coordinate. And finally, Regions are a sorted list of non-intersecting Bands.
"""

from collections.abc import Iterator

from .basic import Point, Pointlike, Size, Sizelike

__all__ = ["Region"]

class Region:
    """
    Collection of mutually exclusive bands of rects.

    Methods
    -------
    rects()
        Yield position and size of rects that make up the region.
    from_rect(pos, size)
        Return a new region from a rect position and size.
    """

    def __and__(self, other: Region) -> Region:
        """Return the intersection of self and other."""

    def __or__(self, other: Region) -> Region:
        """Return the union of self and other."""

    def __add__(self, other: Region) -> Region:
        """Return the union of self and other."""

    def __sub__(self, other: Region) -> Region:
        """Return the difference of self and other."""

    def __xor__(self, other: Region) -> Region:
        """Return the symmetric difference of self and other."""

    def __bool__(self) -> bool:
        """Whether region is non-empty."""

    def __eq__(self, other: object) -> bool:
        """Whether two regions are equal."""

    def __contains__(self, point: Point) -> bool:
        """Return whether point is in region."""

    def rects(self) -> Iterator[tuple[Point, Size]]:
        """
        Yield position and size of rects that make up the region.

        Yields
        ------
        tuple[Point, Size]
            A position and size of a rect in the region.
        """

    @classmethod
    def from_rect(cls, pos: Pointlike, size: Sizelike) -> Region:
        """
        Return a region from a rect position and size.

        Returns
        -------
        Region
            A new region.
        """



"""A movable, resizable window gadget."""

from __future__ import annotations

from typing import Literal

from uwcwidth import wcswidth

from ..terminal.events import MouseEvent
from .behaviors.focusable import Focusable
from .behaviors.grabbable import Grabbable
from .behaviors.themable import Themable
from .gadget import Gadget, Point, Pointlike, PosHint, Size, SizeHint, Sizelike, clamp
from .pane import Pane
from .text import Text

__all__ = ["Point", "Size", "Window"]


class _TitleBar(Grabbable, Pane):
    parent: Window  # type: ignore

    def __init__(self):
        super().__init__(
            pos=(1, 2),
            size=(1, 2),
            size_hint={"width_hint": 1.0, "width_offset": -4},
            is_transparent=False,
        )

        self._label = Text(pos_hint={"x_hint": 0.5, "anchor": "top"})
        self.add_gadget(self._label)

    def grab_update(self, mouse_event):
        self.parent.top += mouse_event.dy
        self.parent.left += mouse_event.dx


class Window(Themable, Focusable, Grabbable, Gadget):
    r"""
    A movable, resizable window gadget.

    The view can be set with the :attr:`view` property, e.g.,
    ``my_window.view = some_gadget``.

    Parameters
    ----------
    title : str, default: ""
        Title of window.
    allow_vertical_resize : bool, default: True
        Whether window can be resized vertically.
    allow_horizontal_resize : bool, default: True
        Whether window can be resized horizontally.
    resize_min_height : int | None, default: None
        Minimum height window can be resized.
    resize_min_width : int | None, default: None
        Minimum width window can be resized.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    view : Gadget
        The windowed gadget.
    title : str
        Title of window.
    is_focused : bool
        Whether gadget has focus.
    any_focused : bool
        Whether any gadget has focus.
    allow_vertical_resize : bool
        Whether window can be resized vertically.
    allow_horizontal_resize : bool
        Whether window can be resized horizontally.
    resize_min_height : int
        Minimum height window can be resized.
    resize_min_width : int
        Minimum width window can be resized.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    focus()
        Focus gadget.
    blur()
        Un-focus gadget.
    focus_next()
        Focus next focusable gadget.
    focus_previous()
        Focus previous focusable gadget.
    on_focus()
        Update gadget when it gains focus.
    on_blur()
        Update gadget when it loses focus.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.

    Notes
    -----
    Windows remove size and pos hints from their children.

    If not given or too small, :attr:`resize_min_height` and
    :attr:`resize_min_width` will be set large enough so that the border and title
    are visible.
    """

    def __init__(
        self,
        *,
        title="",
        allow_vertical_resize: bool = True,
        allow_horizontal_resize: bool = True,
        resize_min_height: int | None = None,
        resize_min_width: int | None = None,
        alpha: float = 1.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_visible: bool = True,
        is_enabled: bool = True,
        is_transparent: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_visible=is_visible,
            is_enabled=is_enabled,
            is_transparent=is_transparent,
        )
        self._y_edge: Literal[-1, 0, 1] = 0
        self._x_edge: Literal[-1, 0, 1] = 0
        self._view = None
        self._titlebar = _TitleBar()
        self._top_border = Pane(size=(1, 1), size_hint={"width_hint": 1.0})
        self._bot_border = Pane(
            size=(1, 1),
            size_hint={"width_hint": 1.0},
            pos_hint={"y_hint": 1.0, "anchor": "bottom"},
        )
        self._left_border = Pane(
            pos=(1, 0),
            size=(1, 2),
            size_hint={"height_hint": 1.0, "height_offset": -2},
        )
        self._right_border = Pane(
            pos=(1, 0),
            size=(1, 2),
            size_hint={"height_hint": 1.0, "height_offset": -2},
            pos_hint={"x_hint": 1.0, "anchor": "right"},
        )
        self._background = Pane(
            pos=(1, 2),
            size_hint={
                "height_hint": 1.0,
                "width_hint": 1.0,
                "height_offset": -2,
                "width_offset": -4,
            },
        )
        self.add_gadgets(
            self._top_border,
            self._bot_border,
            self._left_border,
            self._right_border,
            self._background,
            self._titlebar,
        )
        self._title = ""
        self.allow_vertical_resize = allow_vertical_resize
        self.allow_horizontal_resize = allow_horizontal_resize
        self.resize_min_height = resize_min_height
        self.resize_min_width = resize_min_width
        self.alpha = alpha
        self.title = title

    @property
    def resize_min_height(self) -> int:
        """Minimum height gadget can be resized."""
        return self._resize_min_height

    @resize_min_height.setter
    def resize_min_height(self, min_height: int | None):
        if min_height is None:
            self._resize_min_height = 3
        else:
            self._resize_min_height = clamp(min_height, 3, None)

    @property
    def resize_min_width(self) -> int:
        """Minimum width gadget can be resized."""
        return self._resize_min_width

    @resize_min_width.setter
    def resize_min_width(self, min_width: int | None):
        w = 6 + wcswidth(self._title)
        if min_width is None:
            self._resize_min_width = w
        else:
            self._resize_min_width = clamp(min_width, w, None)

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._background.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._top_border.alpha = alpha
        self._bot_border.alpha = alpha
        self._left_border.alpha = alpha
        self._right_border.alpha = alpha
        self._background.alpha = alpha

    @property
    def view(self) -> Gadget | None:
        """The windowed gadget."""
        return self._view

    @view.setter
    def view(self, view: Gadget | None):
        if self._view is not None:
            self.remove_gadget(self._view)

        self._view = view

        if view is not None:
            view.size_hint = {}
            view.pos_hint = {}
            view.pos = 2, 2
            self.add_gadget(view)
            self.on_size()

    @property
    def title(self):
        """Title of window."""
        return self._title

    @title.setter
    def title(self, title: str):
        self._title = title
        self._titlebar._label.width = wcswidth(title)
        self._titlebar._label.add_str(title)
        self.resize_min_width = self.resize_min_width

    def grab(self, mouse_event: MouseEvent):
        """Grab window."""
        super().grab(mouse_event)
        y, x = self.to_local(mouse_event.pos)
        self._y_edge = -1 if y == 0 else 1 if y == self.height - 1 else 0
        self._x_edge = (
            -1 if x in (0, 1) else 1 if x in (self.width - 2, self.width - 1) else 0
        )
        if self._y_edge == 0 and self._x_edge == 0:
            self.ungrab(mouse_event)

    def grab_update(self, mouse_event: MouseEvent):
        """Resize window if border is grabbed."""
        y, x = self.to_local(mouse_event.pos)
        dy = mouse_event.dy
        dx = mouse_event.dx
        if (
            (dy < 0 and y >= self.height - 1)
            or (dy > 0 and y <= 0)
            or (dx < 0 and x >= self.width - 1)
            or (dx > 0 and x <= 0)
        ):
            return

        verti = self.allow_vertical_resize and self._y_edge
        horiz = self.allow_horizontal_resize and self._x_edge
        h, w = self.size
        new_size = Size(
            clamp(h + verti * dy, self.resize_min_height, None),
            clamp(w + horiz * dx, self.resize_min_width, None),
        )
        self.size = new_size
        if self._y_edge == -1:
            self.top += h - new_size.height
        if self._x_edge == -1:
            self.left += w - new_size.width

    def update_theme(self):
        """Paint the gadget with current theme."""
        self._background.bg_color = self.get_color("primary_bg")

        if self.is_focused:
            fg = self.get_color("titlebar_normal_fg")
            bg = self.get_color("titlebar_normal_bg")
            border = self.get_color("window_border_normal")
        else:
            fg = self.get_color("titlebar_inactive_fg")
            bg = self.get_color("titlebar_inactive_bg")
            border = self.get_color("window_border_inactive")

        self._titlebar.bg_color = bg
        self._titlebar._label.default_fg_color = fg
        self._titlebar._label.default_bg_color = bg
        self._titlebar._label.canvas["fg_color"] = fg
        self._titlebar._label.canvas["bg_color"] = bg
        self._top_border.bg_color = border
        self._bot_border.bg_color = border
        self._left_border.bg_color = border
        self._right_border.bg_color = border

    def on_size(self):
        """Resize view on resize."""
        h, w = self._size
        h -= 3
        w -= 4
        if self._view is not None:
            self._view.size = h, w
            self._view.is_visible = h > 0 and w > 0

    def on_focus(self):
        """Pull window to front and recolor borders on focus."""
        self.update_theme()
        self.pull_to_front()

    def on_blur(self):
        """Recolor borders on blur."""
        self.update_theme()

    def dispatch_mouse(self, mouse_event):
        """Handle any colliding mouse events."""
        return super().dispatch_mouse(mouse_event) or self.collides_point(
            mouse_event.pos
        )



"""
The graphic rendition parameters of a terminal cell (e.g., bold, italic, etc.).

Some time between numpy v2.0.0 and v2.1.0, numpy started casting IntFlag to int64 (See:
https://github.com/numpy/numpy/issues/28199 and https://github.com/numpy/numpy/issues/27540
). This prevents using IntFlag in bitwise operations with uint arrays. So, to remedy
this, this module defines an IntFlag-like enum using numpy dtypes instead. ``Style``
then uses this enum so that bitwise operations on uint arrays work as expected, in
particular, with the ``"style"`` field of ``Text.canvas``.
"""

from enum import KEEP, Flag, ReprEnum
from typing import no_type_check

import numpy as np

__all__ = ["Style", "Uint8Flag"]


def is_single_bit(n: np.integer):
    """Whether a single bit is set in ``n``."""
    if n == 0:
        return False
    return n & (n - 1) == 0


def bit_length(n: np.integer):
    """Return the number of bits necessary to represent ``n`` in binary."""
    return np.ceil(np.log2(n + 1)).astype(type(n))


class NpFlag(Flag, boundary=KEEP):
    """
    Flag-type enum for numpy dtypes.

    Boundary type must be KEEP as support for other boundary types is removed.
    """

    @no_type_check
    @classmethod
    def _missing_(cls, value):
        # Would like for `NpFlag(1)` and similar to work.
        if not isinstance(value, cls._member_type_):
            try:
                value = cls._member_type_(value)
            except (ValueError, OverflowError):  # ? Are there other errors to catch?
                raise ValueError("%r is not a valid %s" % (value, cls.__qualname__))

        # Masks set by EnumType will be 0 since values don't pass isinstance(value, int)
        # checks, so we fix that here.
        if cls._flag_mask_ == 0:
            cls._flag_mask_ = 0
            cls._singles_mask_ = 0
            for member in cls.__members__.values():
                if is_single_bit(member.value):
                    cls._singles_mask_ |= member.value
                cls._flag_mask_ |= member.value
            cls._all_bits_ = 2 ** bit_length(cls._flag_mask_) - 1

        flag_mask = cls._flag_mask_
        singles_mask = cls._singles_mask_
        all_bits = cls._all_bits_
        neg_value = None
        if not ~all_bits <= value <= all_bits or value & (all_bits ^ flag_mask):
            if value < 0:
                value = max(all_bits + 1, 2 ** (bit_length(value))) + value
        if value < 0:
            neg_value = value
            value = all_bits + 1 + value

        unknown = value & ~flag_mask
        aliases = value & ~singles_mask
        member_value = value & singles_mask

        if cls._member_type_ is object:
            pseudo_member = object.__new__(cls)
        else:
            pseudo_member = cls._member_type_.__new__(cls, value)

        if not hasattr(pseudo_member, "_value_"):
            pseudo_member._value_ = value

        if member_value or aliases:
            members = []
            combined_value = 0
            for m in cls._iter_member_(member_value):
                members.append(m)
                combined_value |= m._value_
            if aliases:
                value = member_value | aliases
                for _, pm in cls._member_map_.items():
                    if (
                        pm not in members
                        and pm._value_
                        and pm._value_ & value == pm._value_
                    ):
                        members.append(pm)
                        combined_value |= pm._value_
            unknown = value ^ combined_value
            pseudo_member._name_ = "|".join([m._name_ for m in members])
            if not combined_value:
                pseudo_member._name_ = None
            elif unknown:
                pseudo_member._name_ += f"|{cls._numeric_repr_(unknown)}"
        else:
            pseudo_member._name_ = None

        pseudo_member = cls._value2member_map_.setdefault(value, pseudo_member)
        if neg_value is not None:
            cls._value2member_map_[neg_value] = pseudo_member
        return pseudo_member


class Uint8Flag(np.uint8, ReprEnum, NpFlag):  # type: ignore
    """Integer flag with ``numpy.uint8``."""


class Style(Uint8Flag):
    """The graphic rendition parameters of a terminal cell."""

    NO_STYLE = 0
    BOLD = 0b1
    ITALIC = 0b10
    UNDERLINE = 0b100
    STRIKETHROUGH = 0b1000
    OVERLINE = 0b10000
    REVERSE = 0b100000



"""
A base for creating tree-like views. Tree views are composed of nodes that
can be selected and toggled open or closed.
"""

from __future__ import annotations

from collections.abc import Iterator
from typing import Self, cast

from .behaviors.button_behavior import ButtonBehavior, ButtonState
from .behaviors.themable import Themable
from .gadget import (
    Cell0D,
    Gadget,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    _GadgetList,
)
from .text import Text

__all__ = ["ButtonState", "Point", "Size", "TreeView", "TreeViewNode"]


class TreeViewNode(Themable, ButtonBehavior, Text):
    r"""
    A node of a :class:`TreeView`.

    Parameters
    ----------
    is_leaf : bool, default: True
        Whether node is a leaf node.
    always_release : bool, default: False
        Whether a mouse up event outside the button will trigger it.
    default_cell : Cell0D | str, default: " "
        Default cell of text canvas.
    alpha : float, default: 0.0
        Transparency of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    root_node : TreeViewNode
        Root node of tree.
    always_release : bool
        Whether a mouse up event outside the button will trigger it.
    state : ButtonState
        Current button state. One of `NORMAL`, `HOVER`, `DOWN`.
    canvas : Cell2D
        The array of characters for the gadget.
    chars : Unicode2D
        Return a view of the ords field of the canvas as 1-character unicode strings.
    default_cell : Cell0D
        Default cell of text canvas.
    default_fg_color : Color
        Foreground color of default cell.
    default_bg_color : Color
        Background color of default cell.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    iter_open_nodes()
        Yield all open descendent nodes.
    add_node(node: TreeViewNode)
        Add a child node.
    remove_node(node: TreeViewNode)
        Remove a child node.
    toggle()
        Toggle node open or closed.
    select()
        Select this node.
    unselect()
        Unselect this node.
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    update_normal()
        Paint the normal state.
    update_hover()
        Paint the hover state.
    update_down()
        Paint the down state.
    on_release()
        Triggered when a button is released.
    add_border(style="light", ...)
        Add a border to the gadget.
    add_syntax_highlighting(lexer, style)
        Add syntax highlighting to current text in canvas.
    add_str(str, pos, ...)
        Add a single line of text to the canvas.
    set_text(text, ...)
        Resize gadget to fit text, erase canvas, then fill canvas with text.
    clear()
        Fill canvas with default cell.
    shift(n=1)
        Shift content in canvas up (or down in case of negative `n`).
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        is_leaf=True,
        always_release: bool = False,
        default_cell: Cell0D | str = " ",
        alpha: float = 0.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.is_leaf: bool = is_leaf
        self.is_open: bool = False
        self.is_selected: bool = False
        self.parent_node: Self | None = None
        self.child_nodes: list[Self] = []
        self.tree_view: TreeView | None = None
        self.level: int = -1
        super().__init__(
            always_release=always_release,
            default_cell=default_cell,
            alpha=alpha,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

    def _repaint(self):
        if self.is_selected:
            fg = self.get_color("menu_item_selected_fg")
            bg = self.get_color("menu_item_selected_bg")
        elif self.button_state == "normal":
            fg = self.get_color("primary_fg")
            bg = self.get_color("primary_bg")
        else:
            fg = self.get_color("menu_item_hover_fg")
            bg = self.get_color("menu_item_hover_bg")
        self.canvas["fg_color"] = fg
        self.canvas["bg_color"] = bg

    def on_size(self):
        """Repaint tree on resize."""
        super().on_size()
        self._repaint()

    def update_theme(self):
        """Paint the gadget with current theme."""
        self._repaint()

    def update_normal(self):
        """Paint the normal state."""
        self._repaint()

    def update_hover(self):
        """Paint the hover state."""
        self._repaint()

    def update_down(self):
        """Paint the down state."""
        self._repaint()

    @property
    def root_node(self) -> Self:
        """Root node of tree."""
        if self.parent_node is None:
            return self
        return self.parent_node.root_node

    def iter_open_nodes(self) -> Iterator[Self]:
        """
        Yield all open descendent nodes.

        Yields
        ------
        TreeViewNode
            A descendent open node.
        """
        for child in self.child_nodes:
            yield child

            if child.is_open:
                yield from child.iter_open_nodes()

    def add_node(self, node: Self):
        """
        Add a child node.

        Parameters
        ----------
        node : TreeViewNode
            The node to add.
        """
        self.child_nodes.append(node)

        node.level = self.level + 1
        node.parent_node = self

    def remove_node(self, node: Self):
        """
        Remove a child node.

        Parameters
        ----------
        node : TreeViewNode
            The node to remove.
        """
        self.child_nodes.remove(node)

        node.level = -1
        node.parent_node = None

    def _toggle_update(self):
        """Update state after :meth:`toggle` is called."""

    def toggle(self):
        """Toggle node open or closed."""
        if not self.is_leaf:
            self.is_open = not self.is_open
            self._toggle_update()
            cast(TreeView, self.root_node.tree_view).update_tree_layout()

    def select(self):
        """Select node."""
        tree_view = cast(TreeView, self.root_node.tree_view)
        if tree_view.selected_node is not None:
            tree_view.selected_node.unselect()

        self.is_selected = True
        tree_view.selected_node = self
        self._repaint()

    def unselect(self):
        """Unselect node."""
        self.is_selected = False
        cast(TreeView, self.root_node.tree_view).selected_node = None
        self._repaint()

    def on_release(self):
        """Select and toggle node on release."""
        self.select()
        self.toggle()


class TreeView[T: TreeViewNode](Gadget):
    r"""
    Base for creating tree-like views.

    Parameters
    ----------
    root_node : TreeViewNode
        Root node of tree view.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    root_node : TreeViewNode
        Root node of tree view
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    update_tree_layout()
        Update tree layout after a child node is toggled open or closed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        root_node: T,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.selected_node: T | None = None
        self.children: _GadgetList[T]
        self.root_node = root_node
        root_node.tree_view = self

        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        root_node.toggle()

    def update_tree_layout(self) -> None:
        """Update tree layout after a child node is opened or closed."""



"""
Functions and classes for parsing batgrl markdown.

batgrl's markdown can make text italic, bold, strikethrough, underlined, or overlined.

#### Syntax for batgrl markdown
- italic: `*this is italic text*`
- bold: `**this is bold text**`
- strikethrough: `~~this is strikethrough text~~`
- underlined: `__this is underlined text__`
- overlined: `^^this is overlined text^^`

This module is adapted from `https://github.com/miyuchina/mistletoe/blob/master/mistletoe/core_tokens.py`.
`mistletoe` is licensed under the MIT license.
"""

import sys
from string import punctuation
from typing import Literal, Self
from unicodedata import category

from .geometry import clamp

DELIMITERS = set("*_~^")
WHITESPACE = set(
    " \t\n\x0b\x0c\r\x1c\x1d\x1e\x1f\x85\xa0\u1680\u2000\u2001\u2002\u2003\u2004\u2005"
    "\u2006\u2007\u2008\u2009\u200a\u2028\u2029\u202f\u205f\u3000"
)
PUNCTUATION = set(punctuation) | {
    chr(i) for i in range(sys.maxunicode + 1) if category(chr(i)).startswith("P")
}

MarkdownStyle = Literal["italic", "bold", "strikethrough", "underline", "overline"]


class Delimiter:
    def __init__(self, start: int, end: int, text: str):
        self.delimiter = text[start]
        self.number = end - start
        self.start = start
        self.end = end
        self.open = is_opener(start, end, text)
        self.close = is_closer(start, end, text)

    def match(self, other: Self) -> tuple[int, int, int, bool, bool]:
        nchars = min(self.number, other.number)
        n = clamp(nchars, None, 2)
        self.start += n
        self.number = self.end - self.start

        other.end -= n
        other.number = other.end - other.start
        return n, self.end, other.start, self.number == 0, other.number == 0

    def closed_by(self, other: Self) -> bool:
        if self.delimiter != other.delimiter:
            return False

        if self.delimiter == "*" and (
            self.open and self.close or other.open and other.close
        ):
            return (self.number + other.number) % 3 != 0 or (
                self.number % 3 == 0 and other.number % 3 == 0
            )
        elif self.delimiter != "*":
            return self.number >= 2 and other.number >= 2
        return True


def next_closer(curr_pos: int, delimiters: list[Delimiter]) -> int | None:
    for i in range(curr_pos, len(delimiters)):
        if delimiters[i].close:
            return i
    return None


def matching_opener(
    curr_pos: int, delimiters: list[Delimiter], bottom: int
) -> int | None:
    if curr_pos > 0:
        curr_delimiter = delimiters[curr_pos]
        for i in range(curr_pos - 1, bottom, -1):
            delimiter = delimiters[i]
            if delimiter.open and delimiter.closed_by(curr_delimiter):
                return i
    return None


def preceded_by(start: int, text: str, charset: set[str]):
    preceding_char = text[start - 1] if start > 0 else " "
    return preceding_char in charset


def succeeded_by(end: int, text: str, charset: set[str]):
    succeeding_char = text[end] if end < len(text) else " "
    return succeeding_char in charset


def is_left_delimiter(start: int, end: int, text: str):
    return not succeeded_by(end, text, WHITESPACE) and (
        not succeeded_by(end, text, PUNCTUATION)
        or preceded_by(start, text, PUNCTUATION)
        or preceded_by(start, text, WHITESPACE)
    )


def is_right_delimiter(start: int, end: int, text: str):
    return not preceded_by(start, text, WHITESPACE) and (
        not preceded_by(start, text, PUNCTUATION)
        or succeeded_by(end, text, WHITESPACE)
        or succeeded_by(end, text, PUNCTUATION)
    )


def is_opener(start: int, end: int, text: str) -> bool:
    if text[start] != "_":
        return is_left_delimiter(start, end, text)
    is_right = is_right_delimiter(start, end, text)
    return is_left_delimiter(start, end, text) and (
        not is_right or (is_right and preceded_by(start, text, PUNCTUATION))
    )


def is_closer(start: int, end: int, text: str) -> bool:
    if text[start] != "_":
        return is_right_delimiter(start, end, text)
    is_left = is_left_delimiter(start, end, text)
    return is_right_delimiter(start, end, text) and (
        not is_left or (is_left and succeeded_by(end, text, PUNCTUATION))
    )


def process_emphasis(
    delimiters: list[Delimiter],
) -> list[tuple[int, int, int, int, MarkdownStyle]]:
    curr_pos = 0
    bottoms = dict.fromkeys("*_^~", -1)
    delimiter_to_style = {
        "*": "italic",
        "**": "bold",
        "~~": "strikethrough",
        "__": "underline",
        "^^": "overline",
    }
    matches = []

    while (curr_pos := next_closer(curr_pos, delimiters)) is not None:
        closer = delimiters[curr_pos]
        delimiter = closer.delimiter

        bottom = bottoms[delimiter]

        open_pos = matching_opener(curr_pos, delimiters, bottom)
        if open_pos is not None:
            opener = delimiters[open_pos]
            n, start, end, opener_done, closer_done = opener.match(closer)

            if delimiter == "*" or n == 2:
                matches.append(
                    (
                        n + opener.number,
                        start,
                        end,
                        n + closer.number,
                        delimiter_to_style[delimiter * n],
                    )
                )

            del delimiters[open_pos + 1 : curr_pos]
            curr_pos -= curr_pos - open_pos - 1

            if opener_done:
                delimiters.remove(opener)
                curr_pos -= 1
            if closer_done:
                delimiters.remove(closer)
                curr_pos -= 1
            if curr_pos < 0:
                curr_pos = 0
        else:
            bottom = curr_pos - 1
            bottoms[delimiter] = bottom

            if closer.open:
                curr_pos += 1
            else:
                delimiters.remove(closer)

    return matches


def find_md_tokens(
    text: str,
) -> tuple[list[tuple[int, int, int, int, MarkdownStyle]], list[int]]:
    delimiters = []
    escapes = []
    escaped = False
    delimiter = None
    start = 0
    i = 0
    while i < len(text):
        c = text[i]
        if c == "\\" and not escaped:
            escapes.append(i)
            escaped = True
            i += 1
            continue
        if delimiter is not None and (c != delimiter or escaped):
            delimiters.append(Delimiter(start, i if not escaped else i - 1, text))
            delimiter = None
        if delimiter is None and c in DELIMITERS and not escaped:
            delimiter = c
            start = i
        escaped = False
        i += 1
    if delimiter:
        delimiters.append(Delimiter(start, i, text))
    return process_emphasis(delimiters), escapes



"""A progress bar gadget."""

import asyncio
from itertools import chain, cycle
from typing import cast

from ..text_tools import Style, smooth_horizontal_bar, smooth_vertical_bar
from .behaviors.themable import Themable
from .gadget import (
    Gadget,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    bindable,
    clamp,
)
from .text import Text

__all__ = ["Point", "ProgressBar", "Size"]


class ProgressBar(Themable, Gadget):
    r"""
    A progress bar gadget.

    Setting :attr:`progress` to `None` will start a "loading" animation; otherwise
    setting to a value between `0.0` and `1.0` will update the bar.

    Parameters
    ----------
    animation_delay : float, default: 1/60
        Time between loading animation updates.
    is_horizontal : bool, default: True
        If true, the bar will progress to the right, else upwards.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    progress : float | None
        Current progress as a value between `0.0` and `1.0` or `None`.
    animation_delay : float
        Time between loading animation updates.
    is_horizontal : bool
        If true, the bar will progress to the right, else upwards.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        is_horizontal: bool = True,
        animation_delay: float = 1 / 60,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._bar = Text()
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.add_gadget(self._bar)
        self.animation_delay = animation_delay
        self._is_horizontal = is_horizontal
        self._progress: float | None = None
        self._loading_task: asyncio.Task | None = None

    @property
    def progress(self) -> float | None:
        """
        Current progress as a value between `0.0` and `1.0` or `None`.

        If progress is `None`, a loading animation will play.
        """
        return self._progress

    @progress.setter
    @bindable
    def progress(self, progress: float):
        self._progress = None if progress is None else clamp(progress, 0.0, 1.0)
        self._update_bar()

    def _update_bar(self):
        if self.root is None:
            return

        self._bar.size = self.size

        if self._loading_task is not None:
            self._loading_task.cancel()
            self._loading_task = None

        if self._progress is None:
            self._loading_task = asyncio.create_task(self._loading_animation())
        else:
            self._paint_progress_bar()

    @property
    def is_horizontal(self) -> bool:
        """If true, the bar will progress to the right, else upwards."""
        return self._is_horizontal

    @is_horizontal.setter
    def is_horizontal(self, is_horizontal: bool):
        self._is_horizontal = is_horizontal
        self._update_bar()

    def _paint_small_horizontal_bar(self, progress):
        bar_width = max(1, (self.width - 1) // 2)
        x, offset = divmod(progress * (self.width - bar_width), 1)
        x = int(x)
        smooth_bar = smooth_horizontal_bar(bar_width, 1, offset)

        self._bar.clear()
        self._bar.chars[:, x : x + len(smooth_bar)] = smooth_bar
        self._bar.canvas["fg_color"] = self.get_color("progress_bar_fg")
        self._bar.canvas["bg_color"] = self.get_color("progress_bar_bg")
        if offset:
            self._bar.canvas["style"][:, x] = Style.REVERSE

    def _paint_small_vertical_bar(self, progress):
        bar_height = max(1, (self.height - 1) // 2)
        y, offset = divmod(progress * (self.height - bar_height), 1)
        y = int(y)
        smooth_bar = smooth_vertical_bar(bar_height, 1, offset)

        self._bar.clear()
        self._bar.chars[::-1][y : y + len(smooth_bar)].T[:] = smooth_bar
        self._bar.canvas["fg_color"] = self.get_color("progress_bar_fg")
        self._bar.canvas["bg_color"] = self.get_color("progress_bar_bg")
        if offset:
            self._bar.canvas["style"][::-1][y] = Style.REVERSE

    async def _loading_animation(self):
        if (
            self.is_horizontal
            and self.width < 3
            or not self.is_horizontal
            and self.height < 3
        ):
            return

        self._bar.chars[:] = " "

        if self._is_horizontal:
            steps = 8 * self.width
            paint = self._paint_small_horizontal_bar
        else:
            steps = 8 * self.height
            paint = self._paint_small_vertical_bar

        for i in cycle(chain(range(steps + 1), range(steps)[::-1])):
            paint(i / steps)
            await asyncio.sleep(self.animation_delay)

    def on_add(self):
        """Start loading animation on add if progress is None."""
        super().on_add()
        self._update_bar()

    def on_remove(self):
        """Cancel loading animation on remove."""
        super().on_remove()
        if self._loading_task is not None:
            self._loading_task.cancel()
            self._loading_task = None

    def on_size(self):
        """Repaint bar on resize."""
        self._update_bar()

    def update_theme(self):
        """Paint the gadget with current theme."""
        fg = self.get_color("progress_bar_fg")
        bg = self.get_color("progress_bar_bg")
        self._bar.canvas["fg_color"] = fg
        self._bar.canvas["bg_color"] = bg
        self.default_fg_color = fg
        self.default_bg_color = bg

    def _paint_progress_bar(self):
        self._bar.clear()
        self._bar.canvas["fg_color"] = self.get_color("progress_bar_fg")
        self._bar.canvas["bg_color"] = self.get_color("progress_bar_bg")

        progress = cast(float, self.progress)
        if self.is_horizontal:
            smooth_bar = smooth_horizontal_bar(self.width, progress)
            self._bar.chars[:, : len(smooth_bar)] = smooth_bar
        else:
            smooth_bar = smooth_vertical_bar(self.height, progress)
            self._bar.chars[::-1][: len(smooth_bar)].T[:] = smooth_bar



"""An animation gadget."""

import asyncio
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Self

import numpy as np

from ..array_types import RGBA_2D
from ..colors import TRANSPARENT, AColor
from ..texture_tools import read_texture, resize_texture
from .graphics import (
    Blitter,
    Graphics,
    Interpolation,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    scale_geometry,
)

__all__ = ["Animation", "Interpolation", "Point", "Size"]


class Animation(Graphics):
    r"""
    An animation gadget.

    Parameters
    ----------
    path : Path | None, default: None
        Path to directory of images for frames in the animation (loaded in lexographical
        order of filenames).
    frame_durations : float | Sequence[float], default: 1/12
        Time each frame is displayed. If a sequence is provided, it's length should be
        equal to number of frames.
    loop : bool, default: True
        Whether to restart animation after last frame.
    reverse : bool, default: False
        Whether to play animation in reverse.
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    blitter : Blitter, default: "half"
        Determines how graphics are rendered.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    frames : list[Image]
        Frames of the animation.
    frame_durations : list[float]
        Time each frame is displayed.
    loop : bool
        Whether to animation is restarted after last frame.
    reverse : bool
        Whether to animation is played in reverse.
    texture : RGBA_2D
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    blitter : Blitter
        Determines how graphics are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    play()
        Play the animation. Returns a task.
    pause()
        Pause the animation
    stop()
        Stop the animation and reset current frame.
    from_textures(textures, ...)
        Create an :class:`Animation` from an iterable of uint8 RGBA numpy array.
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        path: Path | None = None,
        frame_durations: float | Sequence[float] = 1 / 12,
        loop: bool = True,
        reverse: bool = False,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.frames: list[RGBA_2D]
        """Frames of the animation."""
        # Set in `on_size`
        self._sized_frames: list[RGBA_2D]
        """Resized frames of the animation."""

        if path is None:
            self.frames = []
        else:
            paths = sorted(path.iterdir(), key=lambda file: file.name)
            self.frames = [read_texture(path) for path in paths]

        self.frame_durations: Sequence[float]
        """Time each frame is displayed."""

        nframes = len(self.frames)
        if isinstance(frame_durations, Sequence):
            if len(frame_durations) != nframes:
                raise ValueError(
                    "number of frames doesn't match number of frame durations"
                )
            self.frame_durations = frame_durations
        else:
            self.frame_durations = [frame_durations] * nframes

        super().__init__(
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.loop = loop
        self.reverse = reverse
        self._i = len(self.frames) - 1 if self.reverse else 0
        self._animation_task = None

    @property
    def texture(self) -> RGBA_2D:
        """uint8 RGBA color array."""
        if self._i < len(self.frames):
            return self._sized_frames[self._i]
        return self._texture

    @texture.setter
    def texture(self, texture: RGBA_2D) -> None:  # type: ignore
        self._texture = texture

    def on_remove(self) -> None:
        """Pause animation."""
        self.pause()
        super().on_remove()

    def on_size(self) -> None:
        """Update size of all frames on resize."""
        size = scale_geometry(self._blitter, self._size)
        self._texture = np.full((*size, 4), self.default_color, np.uint8)
        self._sized_frames = [
            resize_texture(texture, size, self.interpolation) for texture in self.frames
        ]

    async def _play_animation(self):
        while self.frames:
            try:
                await asyncio.sleep(self.frame_durations[self._i])
            except asyncio.CancelledError:
                return

            if self.reverse:
                self._i -= 1
                if self._i < 0:
                    self._i = len(self.frames) - 1
                    if not self.loop:
                        return
            else:
                self._i += 1
                if self._i >= len(self.frames):
                    self._i = 0

                    if not self.loop:
                        return

    def play(self) -> asyncio.Task:
        """
        Play animation.

        Returns
        -------
        asyncio.Task
            The task that plays the animation.
        """
        self.pause()

        if self._i == 0 and self.reverse:
            self._i = len(self.frames) - 1
        elif self._i == len(self.frames) - 1 and not self.reverse:
            self._i = 0

        self._animation_task = asyncio.create_task(self._play_animation())
        return self._animation_task

    def pause(self) -> None:
        """Pause animation."""
        if self._animation_task is not None:
            self._animation_task.cancel()

    def stop(self) -> None:
        """Stop the animation and reset current frame."""
        self.pause()
        self._i = len(self.frames) - 1 if self.reverse else 0

    @classmethod
    def from_textures(
        cls,
        textures: Iterable[RGBA_2D],
        *,
        frame_durations: float | Sequence[float] = 1 / 12,
        loop: bool = True,
        reverse: bool = False,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ) -> Self:
        """
        Create an :class:`Animation` from an iterable of uint8 RGBA numpy array.

        Parameters
        ----------
        textures : Iterable[RGBA_2D]
            An iterable of RGBA textures that will be the frames of the animation.
        frame_durations : float | Sequence[float], default: 1/12
            Time each frame is displayed. If a sequence is provided, it's length should
            be equal to number of frames.
        loop : bool, default: True
            Whether to restart animation after last frame.
        reverse : bool, default: False
            Whether to play animation in reverse.
        default_color : AColor, default: AColor(0, 0, 0, 0)
            Default texture color.
        alpha : float, default: 1.0
            Transparency of gadget.
        interpolation : Interpolation, default: "linear"
            Interpolation used when gadget is resized.
        blitter : Blitter, default: "half"
            Determines how graphics are rendered.
        size : Sizelike, default: Size(10, 10)
            Size of gadget.
        pos : Pointlike, default: Point(0, 0)
            Position of upper-left corner in parent.
        size_hint : SizeHint | None, default: None
            Size as a proportion of parent's height and width.
        pos_hint : PosHint | None, default: None
            Position as a proportion of parent's height and width.
        is_transparent : bool, default: True
            Whether gadget is transparent.
        is_visible : bool, default: True
            Whether gadget is visible. Gadget will still receive input events if not
            visible.
        is_enabled : bool, default: True
            Whether gadget is enabled. A disabled gadget is not painted and doesn't
            receive input events.

        Returns
        -------
        Animation
            A new animation gadget.
        """
        frames = list(textures)
        nframes = len(frames)
        if isinstance(frame_durations, Sequence):
            if len(frame_durations) != nframes:
                raise ValueError(
                    "number of frames doesn't match number of frame durations"
                )
        else:
            frame_durations = [frame_durations] * nframes

        animation = cls(
            loop=loop,
            reverse=reverse,
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        animation.frames = frames
        animation.frame_durations = frame_durations
        animation.on_size()
        return animation



from ._fbuf cimport fbuf
from .events import Event

ctypedef unsigned char uint8
ctypedef unsigned int uint
ctypedef enum ParserState: 
    CSI,
    CSI_PARAMS,
    DECRPM,
    ESCAPE,
    EXECUTE_NEXT,
    GROUND,
    OSC,
    PASTE,


cdef class Vt100Terminal:
    cdef:
        fbuf read_buf, in_buf, out_buf
        ParserState state
        int last_y, last_x
        bint skip_newline
        bint sum_supported
        bint sgr_pixels_mode

    cdef void add_event(Vt100Terminal, Event)
    cdef void feed1(Vt100Terminal, uint8)
    cdef void execute_ansi_escapes(Vt100Terminal)
    cdef void execute_csi(Vt100Terminal)
    cdef void execute_csi_params(Vt100Terminal)
    cdef void execute_mouse(Vt100Terminal, uint*, char)
    cdef void execute_osc(Vt100Terminal)
    cdef void execute_decrpm(Vt100Terminal)
    cdef void dsr_request(Vt100Terminal, bytes)
    cpdef void process_stdin(Vt100Terminal)



"""
Sixel Graphics.

The sixel format involves first creating a palette (generally limited to 256 colors) for
the succeeding band data. Each color in the palette uses the ansi "#i;m;r;g;b" where
``i`` is the color register and ``m`` is the "mode" (``1`` for hsl or ``2`` for rgb).
For mode ``2`` (the only mode batgrl uses), the remaining three parameters ``r``, ``g``,
``b`` are the red, green, and blue color components of the color scaled from 0-100
inclusive.

The remaining data for an image is split into 6-pixel high bands. A six-pixel tall
column in that band is called a sixel. For each band, for every color in the band output
"#i" with ``i`` being a color in the palette followed by color data finally ending with
"$" to return to the start of the band (to output a new color) or "-" to move to the
next band.

For the color data, for each pixel in a sixel with values, ``n``, from 0-5 from top-to-
bottom, if that pixel matches the current color add ``2**n``. The result is a value from
0-63. Add 63 to this result to get a character between "?"-"~". If a character is
repeated, run length encoding, "!rc", may be used instead where ``r`` is the number of
times to repeat ``c``.
"""
# This is nearly verbatim notcurses' sixel.c, but in cython.
# notcurses' sixel.c:
#   https://github.com/dankamongmen/notcurses/blob/master/src/lib/sixel.c
#
# notcurses is copyright 2019-2025 Nick Black et al and is licensed under the Apache
# License, Version 2.0:
#   http://www.apache.org/licenses/LICENSE-2.0
#

from libc.stdio cimport sprintf
from libc.stdlib cimport free, malloc, qsort
from libc.string cimport memset, memcpy
from libc.math cimport round

cimport cython

from ._sixel cimport onode, qnode, qstate
from .terminal._fbuf cimport fbuf, fbuf_printf, fbuf_putn, fbuf_puts

ctypedef unsigned char uint8
ctypedef unsigned int uint

cdef:
    uint QNODES_COUNT = 1000
    uint MAX_COLORS = 256

    class OctTree:
        def __init__(self) -> None:
            if alloc_qstate(&self.qs):
                raise MemoryError

        def __dealloc__(self) -> None:
            free_qstate(&self.qs)

    struct BandExtender:
        size_t length, wrote
        int rle
        char rep

    struct SixelMap:
        size_t nbands
        char ***bands

    struct ActiveColor:
        unsigned int color
        char rep


cdef inline uint8 uint8_to_100(uint8 c):
    cdef uint8 result = <uint8>round(c * 100.0 / 255)
    if result > 99:
        return 99
    return result


cdef inline uint qnode_keys(uint8 *srgb, uint8 *skey):
    skey[0] = <uint8>(
        (((srgb[0] % 10) // 5) << 2)
        + (((srgb[1] % 10) // 5) << 1)
        + ((srgb[2] % 10) // 5)
    )
    return srgb[0] // 10 * 100 + srgb[1] // 10 * 10 + srgb[2] // 10


cdef inline bint is_chosen(const qnode *q):
    return q.cidx & 0x8000


cdef inline uint make_chosen(uint cidx):
    return cidx | 0x8000


cdef inline uint qidx(const qnode *q):
    if q.cidx & 0x8000:
        return q.cidx ^ 0x8000
    return q.cidx


cdef inline int insert_color(qstate *qs, uint8 *srgb):
    cdef:
        uint8 skey, skeynat
        uint key = qnode_keys(srgb, &skey)
        qnode *q
        onode *o

    q = &qs.qnodes[key]
    if q.pop == 0 and q.qlink == 0:
        q.srgb[0] = srgb[0]
        q.srgb[1] = srgb[1]
        q.srgb[2] = srgb[2]
        q.pop = 1
        qs.ncolors += 1
        return 0

    if q.qlink == 0:
        qnode_keys(&q.srgb[0], &skeynat)
        if skey == skeynat:
            q.pop += 1
            return 0

        if qs.onodes_free == 0 or qs.dynnodes_free == 0:
            q.pop += 1
            return 0

        o = qs.onodes + qs.onodes_total - qs.onodes_free
        memset(o, 0, sizeof(onode))
        o.q[skeynat] = &qs.qnodes[QNODES_COUNT + qs.dynnodes_total - qs.dynnodes_free]
        qs.dynnodes_free -= 1
        memcpy(o.q[skeynat], q, sizeof(qnode))
        q.qlink = qs.onodes_total - qs.onodes_free + 1
        qs.onodes_free -= 1
        q.pop = 0
    else:
        o = qs.onodes + q.qlink - 1

    if o.q[skey]:
        o.q[skey].pop += 1
        return 0

    if qs.dynnodes_free == 0:
        return -1

    o.q[skey] = &qs.qnodes[QNODES_COUNT + qs.dynnodes_total - qs.dynnodes_free]
    qs.dynnodes_free -= 1
    o.q[skey].pop = 1
    o.q[skey].srgb[0] = srgb[0]
    o.q[skey].srgb[1] = srgb[1]
    o.q[skey].srgb[2] = srgb[2]
    o.q[skey].qlink = 0
    o.q[skey].cidx = 0
    qs.ncolors += 1
    return 0


cdef inline int find_color(const qstate *qs, uint8 *srgb):
    cdef:
        uint8 skey
        uint key = qnode_keys(srgb, &skey)
        qnode *q = &qs.qnodes[key]

    if q.qlink and q.pop == 0:
        if qs.onodes[q.qlink - 1].q[skey]:
            q = qs.onodes[q.qlink - 1].q[skey]
        else:
            return -1

    return qidx(q)


cdef int qnode_cmp(const void* a, const void* b) noexcept nogil:
    cdef:
        const qnode *qa = <qnode*>a
        const qnode *qb = <qnode*>b
    if qa.pop < qb.pop:
        return -1
    if qa.pop == qb.pop:
        return 0
    return 1


cdef qnode *get_active_set(qstate *qs):
    cdef:
        qnode *active = <qnode*>malloc(sizeof(qnode) * qs.ncolors)
        uint target_idx = 0
        uint total = QNODES_COUNT + qs.dynnodes_total - qs.dynnodes_free
        uint z, s
        onode *o

    if active is NULL:
        return active

    for z in range(total):
        if target_idx >= qs.ncolors:
            break
        if qs.qnodes[z].pop:
            memcpy(&active[target_idx], &qs.qnodes[z], sizeof(qnode))
            active[target_idx].qlink = z
            target_idx += 1
        elif qs.qnodes[z].qlink:
            o = &qs.onodes[qs.qnodes[z].qlink - 1]
            for s in range(8):
                if target_idx >= qs.ncolors:
                    break
                if o.q[s]:
                    memcpy(&active[target_idx], o.q[s], sizeof(qnode))
                    active[target_idx].qlink = <uint>(o.q[s] - qs.qnodes)
                    target_idx += 1
    qsort(active, qs.ncolors, sizeof(qnode), &qnode_cmp)
    return active


cdef inline int find_next_lowest_chosen(
    const qstate *qs, uint z, int i, const qnode **hq
):
    cdef:
        const qnode *h
        const onode *o

    while True:
        h = &qs.qnodes[z]
        if h.pop == 0 and h.qlink:
            o = &qs.onodes[h.qlink - 1]
            while i >= 0:
                h = o.q[i]
                if h and is_chosen(h):
                    hq[0] = h
                    return z * 8 + i
                i += 1
                if i >= 8:
                    break
        elif is_chosen(h):
            hq[0] = h
            return z * 8
        z += 1
        if z >= QNODES_COUNT:
            return -1
        i = 0


cdef inline void choose(
    qstate *qs,
    qnode *q,
    uint z,
    int i,
    int *hi,
    int *lo,
    const qnode **hq,
    const qnode **lq,
):
    cdef int cur
    if not is_chosen(q):
        if <int>(z * 8) > hi[0]:
            hi[0] = find_next_lowest_chosen(qs, z, i, hq)
        cur = z * 8 + (i if i >= 0 else 4)
        if lo[0] == -1:
            q.cidx = qidx(hq[0])
        elif hi[0] == -1 or cur - lo[0] < hi[0] - cur:
            q.cidx = qidx(lq[0])
        else:
            q.cidx = qidx(hq[0])
    else:
        lq[0] = q
        lo[0] = z * 8


cdef inline int merge_color_table(qstate *qs):
    if qs.ncolors == 0:
        return 0

    cdef qnode *qactive = get_active_set(qs)
    if qactive is NULL:
        return -1

    cdef uint cidx = 0, z
    for z in range(qs.ncolors - 1, -1, -1):
        if cidx == MAX_COLORS:
            break
        qs.qnodes[qactive[z].qlink].cidx = make_chosen(cidx)
        cidx += 1
    free(qactive)

    if qs.ncolors <= MAX_COLORS:
        return 0

    cdef:
        int lo = -1, hi = -1, i
        const qnode *lq = NULL
        const qnode *hq = NULL
        const onode *o
    for z in range(QNODES_COUNT):
        if qs.qnodes[z].pop == 0:
            if qs.qnodes[z].qlink == 0:
                continue
            o = &qs.onodes[qs.qnodes[z].qlink - 1]
            for i in range(8):
                if o.q[i]:
                    choose(qs, o.q[i], z, i, &hi, &lo, &hq, &lq)
        else:
            choose(qs, &qs.qnodes[z], z, -1, &hi, &lo, &hq, &lq)
    qs.ncolors = MAX_COLORS
    return 0


cdef inline void load_color_table(const qstate *qs):
    cdef:
        int total = QNODES_COUNT + qs.dynnodes_total - qs.dynnodes_free
        int loaded = 0, z
        const qnode *q

    for z in range(total):
        if loaded == qs.ncolors:
            break
        q = &qs.qnodes[z]
        if is_chosen(q):
            qs.table[3 * qidx(q)] = q.srgb[0]
            qs.table[3 * qidx(q) + 1] = q.srgb[1]
            qs.table[3 * qidx(q) + 2] = q.srgb[2]
            loaded += 1


cdef int alloc_qstate(qstate *qs):
    qs.dynnodes_total = MAX_COLORS
    qs.dynnodes_free = qs.dynnodes_total
    qs.qnodes = <qnode*>malloc((QNODES_COUNT + qs.dynnodes_total) * sizeof(qnode))
    if qs.qnodes is NULL:
        return -1

    qs.onodes_total = qs.dynnodes_total // 8
    qs.onodes_free = qs.onodes_total
    qs.onodes = <onode*>malloc(qs.onodes_total * sizeof(onode))
    if qs.onodes is NULL:
        free(qs.qnodes)
        return -1

    qs.table = <uint8*>malloc(3 * MAX_COLORS)
    if qs.table is NULL:
        free(qs.qnodes)
        free(qs.onodes)
        return -1

    memset(qs.qnodes, 0, sizeof(qnode) * QNODES_COUNT)
    qs.ncolors = 0
    return 0


cdef void reset_qstate(qstate *qs):
    memset(qs.qnodes, 0, sizeof(qnode) * QNODES_COUNT)
    qs.dynnodes_free = qs.dynnodes_total
    qs.onodes_free = qs.onodes_total
    qs.ncolors = 0


cdef void free_qstate(qstate *qs):
    if qs is not NULL:
        free(qs.qnodes)
        free(qs.onodes)
        free(qs.table)


cdef inline size_t sixel_count(size_t h, size_t w):
    return (h + 5) // 6 * w


cdef inline size_t sixel_band_count(size_t h):
    return sixel_count(h, 1)


cdef SixelMap *new_sixel_map(size_t h):
    cdef SixelMap *sixel_map = <SixelMap*>malloc(sizeof(SixelMap))
    if sixel_map is NULL:
        return NULL
    sixel_map.nbands = sixel_band_count(h)
    sixel_map.bands = <char***>malloc(sizeof(char**) * sixel_map.nbands)
    if sixel_map.bands is NULL:
        free(sixel_map)
        return NULL
    return sixel_map


cdef void color_band_free(char **color_band, size_t ncolors):
    cdef size_t i
    for i in range(ncolors):
        if color_band[i] is not NULL:
            free(color_band[i])
    free(color_band)


cdef void sixel_map_free(SixelMap *sixel_map, uint ncolors):
    cdef size_t i
    for i in range(sixel_map.nbands):
        if sixel_map.bands[i] is not NULL:
            color_band_free(sixel_map.bands[i], ncolors)
    free(sixel_map.bands)
    free(sixel_map)


cdef inline void write_rle(char *color, size_t *length, ssize_t rle, char rep):
    if rle > 2:
        length[0] += sprintf(&color[length[0]], "!%d", <int>rle)
    elif rle == 2:
        color[length[0]] = rep
        length[0] += 1
    if rle > 0:
        color[length[0]] = rep
        length[0] += 1
    color[length[0]] = 0


cdef inline char *color_band_extend(
    char *color_band, BandExtender *extender, size_t w, size_t x
):
    if color_band is NULL:
        color_band = <char*>malloc(w + 1)
    if color_band is NULL:
        return NULL
    write_rle(color_band, &extender.length, extender.rle, extender.rep + 63)
    cdef ssize_t clear_len = x - (extender.rle + extender.wrote)
    write_rle(color_band, &extender.length, clear_len, 63)
    return color_band


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int build_sixel_band(
    size_t n,
    SixelMap *sixel_map,
    qstate *qs,
    uint8[:, :, ::1] stexture,
    size_t oy,
    size_t ox,
    size_t h,
    size_t w,
    size_t *P2,
):
    cdef:
        BandExtender *extenders = <BandExtender*>malloc(
            sizeof(BandExtender) * qs.ncolors
        )
        char **color_bands

    if extenders is NULL:
        return -1

    sixel_map.bands[n] = <char**>malloc(sizeof(char*) * qs.ncolors)
    color_bands = sixel_map.bands[n]
    if color_bands is NULL:
        free(extenders)
        return -1

    memset(color_bands, 0, sizeof(char*) * qs.ncolors)
    memset(extenders, 0, sizeof(BandExtender) * qs.ncolors)

    cdef:
        size_t ystart = n * 6, yend = min(ystart + 6, h), x, y, i
        ActiveColor[6] active_colors
        unsigned int cidx, nactive_colors, color
        BandExtender *extender

    for x in range(w):
        nactive_colors = 0
        for y in range(ystart, yend):
            if not stexture[oy + y, ox + x, 3]:
                P2[0] = 1
                continue

            cidx = find_color(qs, &stexture[oy + y, ox + x, 0])
            if cidx < 0:
                return -1
            # Store color index for later use by SEE_THROUGH_SIXEL cells in rendering.
            stexture[oy + y, ox + x, 0] = cidx

            for i in range(nactive_colors):
                if active_colors[i].color == cidx:
                    active_colors[i].rep += 1 << (y - ystart)
                    break
            else:
                active_colors[nactive_colors].color = cidx
                active_colors[nactive_colors].rep = 1 << (y - ystart)
                nactive_colors += 1

        for i in range(nactive_colors):
            color = active_colors[i].color
            extender = &extenders[color]

            if (
                extender.rep == active_colors[i].rep
                and extender.rle + extender.wrote == x
            ):
                extender.rle += 1
            else:
                color_bands[color] = color_band_extend(
                    color_bands[color], extender, w, x
                )
                if color_bands[color] is NULL:
                    free(extenders)
                    return -1
                extender.rle = 1
                extender.wrote = x
                extender.rep = active_colors[i].rep

    for color in range(qs.ncolors):
        extender = &extenders[color]
        if extender.rle == 0:
            color_bands[color] = NULL
        else:
            color_bands[color] = color_band_extend(
                color_bands[color], extender, w, w - 1
            )
            if color_bands[color] is NULL:
                free(extenders)
                return -1

    free(extenders)
    return 0


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int sixel(
    fbuf *f,
    qstate *qs,
    const uint8[:, :, ::1] texture,
    uint8[:, :, ::1] stexture,
    unsigned int aspect_h,
    unsigned int aspect_w,
    size_t oy,
    size_t ox,
    size_t h,
    size_t w,
):
    reset_qstate(qs)

    cdef size_t y, x
    for y in range(oy, oy + h):
        for x in range(ox, ox + w):
            if texture[y, x, 3]:
                stexture[y, x, 0] = uint8_to_100(texture[y, x, 0])
                stexture[y, x, 1] = uint8_to_100(texture[y, x, 1])
                stexture[y, x, 2] = uint8_to_100(texture[y, x, 2])
                stexture[y, x, 3] = 1
                if insert_color(qs, &stexture[y, x, 0]):
                    return -1
            else:
                stexture[y, x, 3] = 0
    if merge_color_table(qs):
        return -1

    load_color_table(qs)

    cdef:
        size_t n, color, P2 = 0
        bint close_previous
        SixelMap *sixel_map = new_sixel_map(h)
        char **color_bands
        uint8 *rgb = qs.table

    if sixel_map is NULL:
        return -1

    for n in range(sixel_map.nbands):
        if build_sixel_band(n, sixel_map, qs, stexture, oy, ox, h, w, &P2) < 0:
            sixel_map_free(sixel_map, qs.ncolors)
            return -1

    if fbuf_printf(f, "\x1bP;%d;q\"%d;%d;%d;%d", P2, aspect_h, aspect_w, w, h):
        sixel_map_free(sixel_map, qs.ncolors)
        return -1

    for color in range(qs.ncolors):
        if fbuf_printf(f, "#%d;2;%d;%d;%d", color, rgb[0], rgb[1], rgb[2]):
            sixel_map_free(sixel_map, qs.ncolors)
            return -1
        rgb += 3

    for n in range(sixel_map.nbands):
        close_previous = 0
        color_bands = sixel_map.bands[n]
        for color in range(qs.ncolors):
            if color_bands[color] is NULL:
                continue
            if close_previous:
                if fbuf_putn(f, "$", 1):
                    sixel_map_free(sixel_map, qs.ncolors)
                    return -1
            else:
                close_previous = 1
            if fbuf_printf(f, "#%d", color):
                sixel_map_free(sixel_map, qs.ncolors)
                return -1
            if fbuf_puts(f, color_bands[color]):
                sixel_map_free(sixel_map, qs.ncolors)
                return -1
        if fbuf_putn(f, "-", 1):
            sixel_map_free(sixel_map, qs.ncolors)
            return -1

    f.len -= 1  # Remove last "-"
    sixel_map_free(sixel_map, qs.ncolors)
    if fbuf_putn(f, "\x1b\\", 2):
        return -1

    return 0



"""Grabbable behavior for a gadget."""

from typing import Literal

from ...terminal.events import MouseButton, MouseEvent
from . import Behavior

__all__ = ["Grabbable"]


class Grabbable(Behavior):
    """
    Grabbable behavior for a gadget.

    Mouse down events that collide with the gadget will "grab" it, calling :meth:`grab`.
    While grabbed, each mouse event will call :meth:`grab_update` until the gadget is
    ungrabbed, i.e., a mouse up event is received (which calls :meth:`ungrab`).

    To customize grabbable behavior, implement any of :meth:`grab`, :meth:`grab_update`,
    or :meth:`ungrab`.

    Parameters
    ----------
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"], default: "left"
        Mouse button used for grabbing.

    Attributes
    ----------
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"]
        Mouse button used for grabbing.
    is_grabbed : bool
        Whether gadget is grabbed.

    Methods
    -------
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    """

    def __init__(
        self,
        *,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton | Literal["any"] = "left",
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.is_grabbable: bool = is_grabbable
        self.ptf_on_grab: bool = ptf_on_grab
        self.mouse_button: MouseButton | Literal["any"] = mouse_button
        self._is_grabbed: bool = False

    def on_mouse(self, mouse_event) -> bool | None:
        """Determine if mouse event grabs or ungrabs gadget."""
        if self.is_grabbable:
            if self.is_grabbed:
                if mouse_event.event_type == "mouse_up":
                    self.ungrab(mouse_event)
                else:
                    self.grab_update(mouse_event)

                return True
            if (
                self.collides_point(mouse_event.pos)
                and mouse_event.event_type == "mouse_down"
                and (
                    self.mouse_button == "any"
                    or mouse_event.button == self.mouse_button
                )
            ):
                self.grab(mouse_event)
                return True

        return super().on_mouse(mouse_event)

    @property
    def is_grabbed(self) -> bool:
        """Whether gadget is grabbed."""
        return self._is_grabbed

    def grab(self, mouse_event: MouseEvent) -> None:
        """
        Grab gadget.

        Parameters
        ----------
        mouse_event : MouseEvent
            The mouse event that grabbed the gadget.
        """
        self._is_grabbed = True

        if self.ptf_on_grab:
            self.pull_to_front()

    def ungrab(self, mouse_event: MouseEvent) -> None:
        """
        Ungrab gadget.

        Parameters
        ----------
        mouse_event : MouseEvent
            The mouse event that ungrabbed the gadget.
        """
        self._is_grabbed = False

    def grab_update(self, mouse_event: MouseEvent) -> None:
        """
        Update grabbed gadget with incoming mouse event.

        Parameters
        ----------
        mouse_event : MouseEvent
            The mouse event that updates the grabbed gadget.
        """



from libc.math cimport round, sqrt
from libc.stdlib cimport free, malloc, realloc
from libc.string cimport memmove

cimport cython

ctypedef unsigned char uint8
ctypedef unsigned long ulong


cdef struct Octant:
    int y
    int x
    bint flipped

# An octant represents an eigth of an area around a light source.
#
# If a light source is at `1`, then an octant with y=-1,x=1,flipped=0 looks like:
#
#  456
#  23
#  1
#
# Where the numbers represent the order cells are visited within an octant. Similarly,
# an octant with y=-1,x=1,flipped=1 looks like:
#
#   6
#  35
# 124
#
# `flipped` determines which axis is moved along first (`y` if not flipped, and `x`
# otherwise). And `y` and `x` determine the quadrant with negative y upwards, and
# negative x leftwards.

cdef Octant[8] OCTANTS
OCTANTS[0].y = 1
OCTANTS[0].x = 1
OCTANTS[0].flipped = 0

OCTANTS[1].y = 1
OCTANTS[1].x = 1
OCTANTS[1].flipped = 1

OCTANTS[2].y = 1
OCTANTS[2].x = -1
OCTANTS[2].flipped = 0

OCTANTS[3].y = 1
OCTANTS[3].x = -1
OCTANTS[3].flipped = 1

OCTANTS[4].y = -1
OCTANTS[4].x = 1
OCTANTS[4].flipped = 0

OCTANTS[5].y = -1
OCTANTS[5].x = 1
OCTANTS[5].flipped = 1

OCTANTS[6].y = -1
OCTANTS[6].x = -1
OCTANTS[6].flipped = 0

OCTANTS[7].y = -1
OCTANTS[7].x = -1
OCTANTS[7].flipped = 1


cdef enum Restrictiveness: PERMISSIVE, MODERATE, RESTRICTIVE


cdef struct Interval:
    double start
    double center
    double end


cdef struct IntervalList:
    Interval *buf
    size_t len, size


cdef size_t INITIAL_INTERVAL_SIZE = 64


cdef int init_intervals(IntervalList *intervals):
    intervals.buf = <Interval*>malloc(INITIAL_INTERVAL_SIZE * sizeof(Interval))
    if intervals.buf is NULL:
        return -1

    intervals.len = 0
    intervals.size = INITIAL_INTERVAL_SIZE


cdef void free_intervals(IntervalList *intervals):
    free(intervals.buf)
    intervals.buf = NULL
    intervals.len = 0
    intervals.size = 0


cdef int grow(IntervalList *intervals, size_t n):
    # Ensure intervals has enough room to grow by `n`.
    if n + intervals.len <= intervals.size:
        return 0

    cdef Interval *new_intervals = <Interval*>realloc(
        intervals.buf, (intervals.size << 1) * sizeof(Interval)
    )
    if new_intervals is NULL:
        return -1

    intervals.buf = new_intervals
    return 0


cdef size_t bisect_intervals(IntervalList *intervals, double value):
    cdef size_t lo = 0, hi = intervals.len, mid
    while lo < hi:
        mid = (lo + hi) // 2
        if value < intervals.buf[mid].start:
            hi = mid
        else:
            lo = mid + 1
    return lo


cdef inline int add_obstruction(IntervalList *obstructions, double start, double end):
    # Insert interval in obstructions while maintaining sorted order.
    cdef:
        size_t a = bisect_intervals(obstructions, start)
        size_t b = bisect_intervals(obstructions, end)

    if a > 0 and start <= obstructions.buf[a - 1].end:
        a -= 1
        start = obstructions.buf[a].start

    if b > 0 and end <= obstructions.buf[b - 1].end:
        end = obstructions.buf[b - 1].end
    elif b < obstructions.len and end == obstructions.buf[b].start:
        b += 1

    if a == b:  # Need to grow intervals
        if grow(obstructions, 1):
            return -1

        if a < obstructions.len:
            memmove(
                &obstructions.buf[a + 1],
                &obstructions.buf[a],
                (obstructions.len - a) * sizeof(Interval),
            )
        obstructions.len += 1
    elif b - 1 - a:  # Need to shrink intervals
        memmove(
            &obstructions.buf[a + 1],
            &obstructions.buf[b],
            (obstructions.len - b) * sizeof(Interval),
        )
        obstructions.len -= b - 1 - a

    obstructions.buf[a].start = start
    obstructions.buf[a].end = end
    obstructions.buf[a].center = (start + end) / 2
    return 0


cdef inline double clamp(double a, double lo, double hi):
    if a < lo:
        return lo
    if a > hi:
        return hi
    return a


cdef inline double dist(double ay, double ax, double by, double bx):
    cdef double y = ay - by, x = ax - bx
    return sqrt(y * y + x * x)


cdef inline double clight_decay(double distance, object light_decay):
    # Need a typed cdef to wrap callable `light_decay`
    return light_decay(distance)


cdef inline bint point_is_visible(
    IntervalList *obstructions,
    double start,
    double end,
    double center,
    Restrictiveness restrict,
):
    cdef:
        bint start_visible = 1, center_visible = 1, end_visible = 1
        size_t a = bisect_intervals(obstructions, start)
        size_t b = bisect_intervals(obstructions, end)
        size_t i
        Interval *obstruction

    if a > 0 and start <= obstructions.buf[a - 1].end:
        a -= 1
    if b < obstructions.len and end == obstructions.buf[b].start:
        end_visible = 1

    for i in range(a, b):
        obstruction = &obstructions.buf[i]
        if start_visible and obstruction.start <= start <= obstruction.end:
            start_visible = 0
        if center_visible and obstruction.start <= center <= obstruction.end:
            center_visible = 0
        if end_visible and obstruction.start <= end <= obstruction.end:
            end_visible = 0

    if restrict == PERMISSIVE:
        return center_visible | start_visible | end_visible
    if restrict == MODERATE:
        return center_visible & (start_visible | end_visible)
    if restrict == RESTRICTIVE:
        return center_visible & start_visible & end_visible


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void cast_shadows(
    uint8[:, :, ::1] texture,
    double[:, :, ::1] light_intensity,
    ulong[:, ::1] caster_map,
    tuple[int, int] camera_pos,
    tuple[int, int] camera_size,
    list[tuple[int, int, int, int]] tile_colors,
    list[tuple[double, double]] light_coords,
    list[tuple[int, int, int]] light_colors,
    str restrictiveness,
    unsigned int radius,
    double smoothing,
    bint not_visible_blocks,
    object light_decay,
):
    cdef Restrictiveness restrict

    if restrictiveness == "permissive":
        restrict = PERMISSIVE
    elif restrictiveness == "moderate":
        restrict = MODERATE
    elif restrictiveness == "restrictive":
        restrict = RESTRICTIVE
    else:
        restrict = PERMISSIVE

    cdef size_t h = texture.shape[0], w = texture.shape[1]
    if h == 0 or w == 0:
        return

    cdef size_t camera_h = camera_size[0], camera_w = camera_size[1]
    if camera_h == 0 or camera_w == 0:
        return

    cdef size_t map_h = caster_map.shape[0], map_w = caster_map.shape[1]
    if map_h == 0 or map_w == 0:
        return

    cdef IntervalList obstructions
    if init_intervals(&obstructions):
        raise MemoryError

    cdef:
        size_t i, j
        int k, m
        size_t nlights = len(light_coords)
        int origin_y, origin_x, y, x, offset_y, offset_x
        int camera_y = camera_pos[0], camera_x = camera_pos[1]
        double scale_y = camera_h / h, scale_x = camera_w / w
        double theta, distance, decay, start, end, center, map_y, map_x
        double smooth_radius = <double>radius + smoothing
        int scaled_radius
        double light_y, light_x
        uint8 r, g, b
        Octant *octant

    for i in range(nlights):
        light_y = light_coords[i][0]
        light_x = light_coords[i][1]

        r = light_colors[i][0]
        g = light_colors[i][1]
        b = light_colors[i][2]

        origin_y = <int>round((light_y - camera_y) / scale_y)
        origin_x = <int>round((light_x - camera_x) / scale_x)

        # Since the octants will be spread around the origin, the origin is handled
        # separately.
        if 0 <= origin_y < h and 0 <= origin_x < w:
            decay = clight_decay(0, light_decay)
            light_intensity[origin_y, origin_x, 0] += r * decay
            light_intensity[origin_y, origin_x, 1] += g * decay
            light_intensity[origin_y, origin_x, 2] += b * decay

        for j in range(8):
            obstructions.len = 0
            octant = &OCTANTS[j]

            # Spread octants around origin so they don't intersect using these offsets.
            # This caster doesn't just mark cells visible or not visible, it
            # determines the brightness of a cell by the distance from the light
            # source. If octants intersected, bright lines would show along their
            # intersection.
            offset_y = octant.y
            offset_x = octant.x
            if offset_y * offset_x > 0:
                offset_y *= 1 - octant.flipped
            else:
                offset_x *= octant.flipped
            if octant.flipped:
                scaled_radius = <int>(radius / scale_x)
            else:
                scaled_radius = <int>(radius / scale_y)

            for k in range(scaled_radius):
                if (
                    obstructions.len == 1
                    and obstructions.buf[0].start == 0.0
                    and obstructions.buf[0].end == 1.0
                ):
                    break

                if octant.flipped:
                    x = origin_x + offset_x + k * octant.x
                else:
                    y = origin_y + offset_y + k * octant.y

                theta = 1.0 / <double>(k + 1)

                for m in range(k + 1):
                    if octant.flipped:
                        y = origin_y + offset_y + m * octant.y
                    else:
                        x = origin_x + offset_x + m * octant.x

                    if y < 0 or y >= h or x < 0 or x >= w:
                        continue

                    map_y = scale_y * y + camera_y
                    map_x = scale_x * x + camera_x
                    distance = dist(light_y, light_x, map_y, map_x)

                    if distance > smooth_radius:
                        continue

                    start = m * theta
                    end = start + theta
                    center = (start + end) / 2

                    if point_is_visible(&obstructions, start, end, center, restrict):
                        decay = clight_decay(distance, light_decay)
                        light_intensity[y, x, 0] += r * decay
                        light_intensity[y, x, 1] += g * decay
                        light_intensity[y, x, 2] += b * decay

                        if map_y < 0 or map_y >= map_h or map_x < 0 or map_x >= map_w:
                            continue
                        if caster_map[<int>map_y, <int>map_x] != 0:
                            if add_obstruction(&obstructions, start, end):
                                raise MemoryError
                    elif not_visible_blocks:
                        if add_obstruction(&obstructions, start, end):
                            raise MemoryError

    free_intervals(&obstructions)

    cdef:
        double intensity_r, intensity_g, intensity_b
        uint8 a
        tuple[int, int, int, int] tile_color

    for y in range(h):
        for x in range(w):
            map_y = y * scale_y + camera_y
            map_x = x * scale_x + camera_x
            if 0 <= map_y < map_h and 0 <= map_x < map_w:
                tile_color = tile_colors[caster_map[<int>map_y, <int>map_x]]
                r = tile_color[0]
                g = tile_color[1]
                b = tile_color[2]
                a = tile_color[3]
            else:
                r = tile_colors[0][0]
                g = tile_colors[0][1]
                b = tile_colors[0][2]
                a = tile_colors[0][3]

            intensity_r = clamp(light_intensity[y, x, 0] / 255, 0.0, 1.0)
            intensity_g = clamp(light_intensity[y, x, 1] / 255, 0.0, 1.0)
            intensity_b = clamp(light_intensity[y, x, 2] / 255, 0.0, 1.0)

            texture[y, x, 0] = <uint8>(r * intensity_r)
            texture[y, x, 1] = <uint8>(g * intensity_g)
            texture[y, x, 2] = <uint8>(b * intensity_b)
            texture[y, x, 3] = a



from collections.abc import Iterator

from libc.stdlib cimport malloc, free, realloc
from libc.string cimport memset

from .basic import Point, Size
from .regions cimport Band, CRegion, Region


cdef Band EMPTY
memset(&EMPTY, 0, sizeof(Band))


ctypedef bint (*bool_op)(bint, bint)


cdef bint cor(bint a, bint b):
    return a | b


cdef bint cand(bint a, bint b):
    return a & b


cdef bint cxor(bint a, bint b):
    return a ^ b


cdef bint csub(bint a, bint b):
    return a & (1 - b)


cdef inline int add_wall(Band *band, int wall):
    cdef int *new_walls
    if band.len == band.size:
        new_walls = <int*>realloc(band.walls, sizeof(Band) * (band.size << 1))
        if new_walls is NULL:
            return -1
        band.size <<= 1
        band.walls = new_walls

    band.walls[band.len] = wall
    band.len += 1
    return 0


cdef inline int add_band(CRegion *region):
    cdef size_t i
    if region.len == region.size:
        new_bands = <Band*>realloc(region.bands, sizeof(Band) * (region.size << 1))
        if new_bands is NULL:
            return -1
        for i in range(region.size, region.size << 1):
            new_bands[i].walls = NULL
        region.size <<= 1
        region.bands = new_bands

    cdef Band *new_band = &region.bands[region.len]
    if new_band.walls is NULL:
        new_band.walls = <int*>malloc(sizeof(int) * 8)
        if new_band.walls is NULL:
            return -1
        new_band.size = 8
    new_band.len = 0

    region.len += 1
    return 0


cdef int merge_bands(
    int y1, int y2, Band *r, Band *s, CRegion *region, bool_op op
):
    if add_band(region) == -1:
        return -1

    cdef:
        Band *new_band = &region.bands[region.len - 1]
        size_t i = 0, j = 0
        bint inside_r = 0, inside_s = 0, inside_region = 0
        int threshold

    while i < r.len or j < s.len:
        if i >= r.len:
            threshold = s.walls[j]
            inside_s ^= 1
            j += 1
        elif j >= s.len:
            threshold = r.walls[i]
            inside_r ^= 1
            i += 1
        elif r.walls[i] < s.walls[j]:
            threshold = r.walls[i]
            inside_r ^= 1
            i += 1
        elif s.walls[j] < r.walls[i]:
            threshold = s.walls[j]
            inside_s ^= 1
            j += 1
        else:
            threshold = r.walls[i]
            inside_r ^= 1
            inside_s ^= 1
            i += 1
            j += 1

        if op(inside_r, inside_s) != inside_region:
            inside_region ^= 1
            if add_wall(new_band, threshold) == -1:
                return -1

    if new_band.len == 0:
        region.len -= 1
        return 0

    new_band.y1 = y1
    new_band.y2 = y2

    if region.len < 2:
        return 0

    cdef Band *previous = &region.bands[region.len - 2]
    if previous.y2 < new_band.y1 or previous.len != new_band.len:
        return 0

    for i in range(previous.len):
        if previous.walls[i] != new_band.walls[i]:
            return 0

    # All walls equal, extend previous band and delete new band.
    previous.y2 = new_band.y2
    region.len -= 1
    return 0


cdef int merge_regions(CRegion *a, CRegion *b, CRegion *result, bool_op op):
    cdef:
        Band *r
        Band *s
        unsigned int i = 0, j = 0
        int scanline = 0

    if a.len > 0:
        if b.len > 0:
            if a.bands[0].y1 < b.bands[0].y1:
                scanline = a.bands[0].y1
            else:
                scanline = b.bands[0].y1
        else:
            scanline = a.bands[0].y1
    elif b.len > 0:
        scanline = b.bands[0].y1

    while i < a.len and j < b.len:
        r = &a.bands[i]
        s = &b.bands[j]

        if r.y1 <= s.y1:
            if scanline < r.y1:
                scanline = r.y1
            if r.y2 <= s.y1:
                if merge_bands(scanline, r.y2, r, &EMPTY, result, op) == -1:
                    return -1
                scanline = r.y2
                i += 1
            else:
                if scanline < s.y1:
                    if merge_bands(scanline, s.y1, r, &EMPTY, result, op) == -1:
                        return -1
                    scanline = s.y1
                if r.y2 <= s.y2:
                    if merge_bands(s.y1, r.y2, r, s, result, op) == -1:
                        return -1
                    scanline = r.y2
                    i += 1
                    if r.y2 == s.y2:
                        j += 1
                else:
                    if merge_bands(s.y1, s.y2, r, s, result, op) == -1:
                        return -1
                    scanline = s.y2
                    j += 1
        else:
            if scanline < s.y1:
                scanline = s.y1
            if s.y2 <= r.y1:
                if merge_bands(scanline, s.y2, &EMPTY, s, result, op) == -1:
                    return -1
                scanline = s.y2
                j += 1
            else:
                if scanline < r.y1:
                    if merge_bands(scanline, r.y1, &EMPTY, s, result, op) == -1:
                        return -1
                    scanline = r.y1
                if s.y2 <= r.y2:
                    if merge_bands(r.y1, s.y2, r, s, result, op) == -1:
                        return -1
                    scanline = s.y2
                    j += 1
                    if s.y2 == r.y2:
                        i += 1
                else:
                    if merge_bands(r.y1, r.y2, r, s, result, op) == -1:
                        return -1
                    scanline = r.y2
                    i += 1

    while i < a.len:
        r = &a.bands[i]
        if scanline < r.y1:
            scanline = r.y1
        if merge_bands(scanline, r.y2, r, &EMPTY, result, op) == -1:
            return -1
        i += 1

    while j < b.len:
        s = &b.bands[j]
        if scanline < s.y1:
            scanline = s.y1
        if merge_bands(scanline, s.y2, &EMPTY, s, result, op) == -1:
            return -1
        j += 1

    return 0


cdef inline size_t bisect_bands(CRegion *region, int y):
    cdef size_t lo = 0, hi = region.len, mid
    while lo < hi:
        mid = (lo + hi) // 2
        if y < region.bands[mid].y1:
            hi = mid
        else:
            lo = mid + 1
    return lo


cdef inline size_t bisect_walls(Band *band, int x):
    cdef size_t lo = 0, hi = band.len, mid
    while lo < hi:
        mid = (lo + hi) // 2
        if x < band.walls[mid]:
            hi = mid
        else:
            lo = mid + 1
    return lo


cdef bint contains(CRegion *cregion, int y, int x):
    cdef size_t i

    i = bisect_bands(cregion, y)
    if i == 0:
        return 0

    if cregion.bands[i - 1].y2 <= y:
        return 0

    i = bisect_walls(&cregion.bands[i - 1], x)
    return i % 2 == 1


cdef void bounding_rect(CRegion *cregion, int *y, int *x, size_t *h, size_t *w):
    if not cregion.len:
        return

    cdef:
        size_t i
        int min_x = cregion.bands[0].walls[0], max_x = min_x
        Band *band

    for i in range(cregion.len):
        band = &cregion.bands[i]
        if band.walls[0] < min_x:
            min_x = band.walls[0]
        if band.walls[band.len - 1] > max_x:
            max_x = band.walls[band.len - 1]

    y[0] = cregion.bands[0].y1
    h[0] = cregion.bands[cregion.len - 1].y2 - y[0]
    x[0] = min_x
    w[0] = max_x - min_x


cdef class Region:
    def __cinit__(self):
        self.cregion.bands = <Band*>malloc(sizeof(Band) * 8)
        if self.cregion.bands is NULL:
            raise MemoryError
        cdef int i
        for i in range(8):
            self.cregion.bands[i].walls = NULL
        self.cregion.size = 8
        self.cregion.len = 0

    def __dealloc__(self):
        if self.cregion.bands is NULL:
            return

        cdef size_t i
        for i in range(self.cregion.len):
            if self.cregion.bands[i].walls is not NULL:
                free(self.cregion.bands[i].walls)
                self.cregion.bands[i].walls = NULL
        free(self.cregion.bands)
        self.cregion.bands = NULL

    @classmethod
    def from_rect(cls, pos: Point, size: Size) -> Region:
        out = Region()
        if size[0] == 0 or size[1] == 0:
            return out

        if add_band(&out.cregion) == -1:
            raise MemoryError

        cdef:
            int y, x, h, w
            Band *band

        y, x = pos
        h, w = size
        band = &out.cregion.bands[0]
        band.y1 = y
        band.y2 = y + h
        band.walls[0] = x
        band.walls[1] = x + w
        band.len = 2

        return out

    def __str__(self) -> str:
        cdef Band *band
        band_reprs = []
        for i in range(self.cregion.len):
            band = &self.cregion.bands[i]
            walls = [band.walls[i] for i in range(band.len)]
            band_reprs.append(f"Band(y1={band.y1}, y2={band.y2}, walls={walls})")
        return f"Region(bands=[{', '.join(band_reprs)}])"

    def __and__(self, other: Region) -> Region:
        out = Region()
        if merge_regions(&self.cregion, &other.cregion, &out.cregion, cand) == -1:
            raise MemoryError
        return out

    def __or__(self, other: Region) -> Region:
        out = Region()
        if merge_regions(&self.cregion, &other.cregion, &out.cregion, cor) == -1:
            raise MemoryError
        return out

    def __add__(self, other: Region) -> Region:
        out = Region()
        if merge_regions(&self.cregion, &other.cregion, &out.cregion, cor) == -1:
            raise MemoryError
        return out

    def __sub__(self, other: Region) -> Region:
        out = Region()
        if merge_regions(&self.cregion, &other.cregion, &out.cregion, csub) == -1:
            raise MemoryError
        return out

    def __xor__(self, other: Region) -> Region:
        out = Region()
        if merge_regions(&self.cregion, &other.cregion, &out.cregion, cxor) == -1:
            raise MemoryError
        return out

    def __bool__(self) -> bool:
        return self.cregion.len > 0

    def __eq__(self, other: Region) -> bool:
        if not isinstance(other, Region):
            return False
        
        cdef:
            size_t i, j
            Band *r
            Band *s
        if self.cregion.len != other.cregion.len:
            return False
        for i in range(self.cregion.len):
            r = &self.cregion.bands[i]
            s = &other.cregion.bands[i]
            if r.y1 != s.y1 or r.y2 != s.y2 or r.len != s.len:
                return False
            for j in range(r.len):
                if r.walls[j] != s.walls[j]:
                    return False
        return True

    def __contains__(self, point: Point) -> bool:
        cdef int y = point[0], x = point[1]
        return contains(&self.cregion, y, x)

    def rects(self) -> Iterator[tuple[Point, Size]]:
        cdef:
            size_t i, j
            Band *band

        for i in range(self.cregion.len):
            band = &self.cregion.bands[i]
            j = 0
            while j < band.len:
                yield (
                    Point(band.y1, band.walls[j]),
                    Size(band.y2 - band.y1, band.walls[j + 1] - band.walls[j]),
                )
                j += 2
